import datetime

from bson import ObjectId
from django.test import TestCase

from django_mongodb_extensions.mql_panel.forms import MQLQueryForm


class MQLPanelFormTests(TestCase):
    def setUp(self):
        self.form = MQLQueryForm()

    def test_empty_documents(self):
        rows, headers = self.form.convert_documents_to_table([])
        self.assertEqual(rows, [])
        self.assertEqual(headers, [])

    def test_simple_fields(self):
        cases = [
            ("username", {"value": "username", "is_json": False}),
            (42, {"value": "42", "is_json": False}),
            (3.14, {"value": "3.14", "is_json": False}),
            (True, {"value": "True", "is_json": False}),
        ]
        for input_value, expected in cases:
            with self.subTest(value=input_value):
                self.assertEqual(self.form._format_cell_value(input_value), expected)

    def test_objectid(self):
        oid = ObjectId("69cd52da72ad0703d3dfef51")
        result = self.form._format_cell_value(oid)
        self.assertIs(result["is_json"], False)
        self.assertEqual(result["value"], "69cd52da72ad0703d3dfef51")

    def test_datetime(self):
        dt = datetime.datetime(2024, 1, 1, 12, 30)
        result = self.form._format_cell_value(dt)
        self.assertIs(result["is_json"], False)
        self.assertEqual(result["value"], "2024-01-01T12:30:00Z")

    def test_embedded_document(self):
        embedded_doc = {
            "name": "Alice",
            "age": 30,
            "created": datetime.datetime(2024, 2, 1, 18, 0),
        }
        result = self.form._format_cell_value(embedded_doc)
        self.assertEqual(result["type"], "dict")
        self.assertIs(result["is_json"], False)
        value_map = {item["key"]: item["value"] for item in result["value"]}
        self.assertEqual(value_map["name"], "Alice")
        self.assertEqual(value_map["age"], "30")
        self.assertEqual(value_map["created"], "2024-02-01T18:00:00Z")
        for item in result["value"]:
            self.assertIs(item["is_json"], False)

    def test_list_field(self):
        arr = ["tag1", "tag2", ObjectId("69cd51ddf1a98c14c906c51e")]
        result = self.form._format_cell_value(arr)
        self.assertEqual(result["type"], "list")
        self.assertIs(result["is_json"], False)
        value_map = {item["key"]: item["value"] for item in result["value"]}
        self.assertEqual(value_map[0], "tag1")
        self.assertEqual(value_map[1], "tag2")
        self.assertEqual(value_map[2], "69cd51ddf1a98c14c906c51e")
        for item in result["value"]:
            self.assertIs(item["is_json"], False)

    def test_list_of_embedded_dicts(self):
        tags = [
            {"name": "cool_tag", "number": 42},
            {"name": "other_tag", "number": 7},
        ]
        result = self.form._format_cell_value(tags)
        self.assertEqual(result["type"], "list")
        self.assertIs(result["is_json"], False)
        for item in result["value"]:
            self.assertEqual(item["type"], "dict")
            self.assertIs(item["is_json"], False)
            inner_map = {inner["key"]: inner["value"] for inner in item["value"]}
            self.assertIn("name", inner_map)
            self.assertIn("number", inner_map)

    def test_dict_with_nested_list_of_dicts(self):
        address = {
            "street": "123 Main St",
            "tags": [
                {"name": "cool_tag", "number": 42},
            ],
        }
        result = self.form._format_cell_value(address)
        self.assertEqual(result["type"], "dict")
        self.assertIs(result["is_json"], False)
        items_by_key = {item["key"]: item for item in result["value"]}
        self.assertEqual(items_by_key["street"]["value"], "123 Main St")
        tags_item = items_by_key["tags"]
        self.assertEqual(tags_item["type"], "list")
        self.assertIs(tags_item["is_json"], False)
        first_tag = tags_item["value"][0]
        self.assertEqual(first_tag["type"], "dict")
        inner_map = {inner["key"]: inner["value"] for inner in first_tag["value"]}
        self.assertEqual(inner_map["name"], "cool_tag")
        self.assertEqual(inner_map["number"], "42")

    def test_deeply_nested_embedded_document(self):
        address = {
            "street": "123 Main St",
            "tags": [
                {
                    "name": "cool_tag",
                    "number": 42,
                    "person": {"name": "Alice"},
                }
            ],
        }
        result = self.form._format_cell_value(address)
        self.assertEqual(result["type"], "dict")
        address_by_key = {item["key"]: item for item in result["value"]}
        tags_item = address_by_key["tags"]
        self.assertEqual(tags_item["type"], "list")
        self.assertIs(tags_item["is_json"], False)
        first_tag = tags_item["value"][0]
        self.assertEqual(first_tag["type"], "dict")
        self.assertIs(first_tag["is_json"], False)
        tag_by_key = {inner["key"]: inner for inner in first_tag["value"]}
        self.assertEqual(tag_by_key["name"]["value"], "cool_tag")
        self.assertEqual(tag_by_key["number"]["value"], "42")
        person_item = tag_by_key["person"]
        self.assertEqual(person_item["type"], "dict")
        self.assertIs(person_item["is_json"], False)
        person_map = {inner["key"]: inner["value"] for inner in person_item["value"]}
        self.assertEqual(person_map["name"], "Alice")



import uuid
from collections import defaultdict

from debug_toolbar.forms import SignedDataForm
from debug_toolbar.panels.sql.forms import SQLSelectForm
from debug_toolbar.panels.sql.panel import SQLPanel
from debug_toolbar.panels.sql.utils import contrasting_color_generator
from debug_toolbar.utils import render_stacktrace
from django.db import connections
from django.db.backends.signals import connection_created
from django.template.loader import render_to_string
from django.urls import path
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from . import views
from .utils import (
    get_mql_warning_threshold,
    patch_get_collection,
    patch_new_connection,
)

connection_created.connect(
    patch_new_connection,
    dispatch_uid="django_mongodb_extensions_mql_panel_patch_new_connection",
)


class MQLPanel(SQLPanel):
    nav_title = _("MQL")
    template = "mql_panel/mql.html"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mql_time = 0
        self._queries = []
        self._databases = {}

    @staticmethod
    def _hex_to_rgb(hex_color):
        """Convert a hex color string to RGB values.

        Used to convert hex colors from contrasting_color_generator() to RGB
        format for display in the debug toolbar UI.
        """
        hex_color = hex_color.lstrip("#")
        if len(hex_color) != 6:
            # Return a default gray color if invalid
            return [128, 128, 128]
        try:
            # Convert hex to RGB
            return [int(hex_color[i : i + 2], 16) for i in (0, 2, 4)]
        except ValueError:
            return [128, 128, 128]

    @staticmethod
    def _is_read_operation(operation):
        return operation in {"aggregate"}

    @cached_property
    def content(self):
        stats = self.get_stats()
        colors = contrasting_color_generator()
        trace_colors = defaultdict(lambda: next(colors))
        for query in stats.get("queries", []):
            query["mql"] = query.get("mql", "")
            query["params"] = True
            query["form"] = SignedDataForm(
                auto_id=None,
                initial=SQLSelectForm(
                    initial={
                        "djdt_query_id": query["djdt_query_id"],
                        "request_id": self.toolbar.request_id,
                    }
                ).initial,
            )
            query["stacktrace"] = render_stacktrace(query["stacktrace"])
            query["trace_color"] = trace_colors[query["stacktrace"]]
        return render_to_string(self.template, stats)

    def disable_instrumentation(self):
        for connection in connections.all():
            if hasattr(connection, "_mql_djdt_logger"):
                connection._mql_djdt_logger = None

    def enable_instrumentation(self):
        # Only patch MongoDB connections (those with get_collection method).
        # This allows the panel to work in multi-database setups with
        # both MongoDB and relational databases.
        #
        # Use _mql_djdt_logger (not _djdt_logger) to avoid conflicting with the
        # SQL panel, which sets _djdt_logger on all connections regardless of type.
        for connection in connections.all():
            if hasattr(connection, "get_collection"):
                patch_get_collection(connection)
                connection._mql_djdt_logger = self

    def generate_stats(self, request, response):
        duplicate_query_groups = defaultdict(list)
        if self._queries:
            mql_warning_threshold = get_mql_warning_threshold()
            db_colors = contrasting_color_generator()
            for db in self._databases.values():
                hex_color = next(db_colors)
                db["rgb_color"] = self._hex_to_rgb(hex_color)
            width_ratio_tally = 0
            for query in self._queries:
                alias = query["alias"]
                dup_key = query.get("mql", "")
                duplicate_query_groups[(alias, dup_key)].append(query)
                query["is_slow"] = query["duration"] > mql_warning_threshold
                operation = query.get("mql_operation", "")
                # Only show Query/Explain buttons if it's a read operation and
                # the args were successfully serialized (mql_args_json is not
                # None).
                args_json = query.get("mql_args_json")
                query["is_query"] = (
                    self._is_read_operation(operation) and args_json is not None
                )
                query["rgb_color"] = self._databases[alias]["rgb_color"]
                try:
                    query["width_ratio"] = (query["duration"] / self._mql_time) * 100
                except ZeroDivisionError:
                    query["width_ratio"] = 0
                query["start_offset"] = width_ratio_tally
                query["end_offset"] = query["width_ratio"] + query["start_offset"]
                width_ratio_tally += query["width_ratio"]
        duplicate_colors = contrasting_color_generator()
        duplicate_counts = defaultdict(int)
        for (alias, _key), query_group in duplicate_query_groups.items():
            count = len(query_group)
            # Duplicates only apply when there are at least 2 identical queries.
            if count > 1:
                color = next(duplicate_colors)
                for query in query_group:
                    query["duplicate_count"] = count
                    query["duplicate_color"] = color
                duplicate_counts[alias] += count
        for alias, db_info in self._databases.items():
            db_info["duplicate_count"] = duplicate_counts[alias]
        self.record_stats(
            {
                "databases": sorted(
                    self._databases.items(), key=lambda x: -x[1]["time_spent"]
                ),
                "queries": self._queries,
                "mql_time": self._mql_time,
            }
        )

    def generate_server_timing(self, request, response):
        stats = self.get_stats()
        num_queries = len(stats.get("queries", []))
        title = f"MQL {num_queries} queries"
        value = stats.get("mql_time", 0)
        self.record_server_timing("mql_time", title, value)

    @classmethod
    def get_urls(cls):
        return [
            path("mql_query/", views.mql_query, name="mql_query"),
            path("mql_explain/", views.mql_explain, name="mql_explain"),
        ]

    @property
    def has_content(self):
        return bool(self._queries)

    @property
    def nav_subtitle(self):
        stats = self.get_stats()
        query_count = len(stats.get("queries", []))
        return ngettext(
            "%(query_count)d query in %(mql_time).3f ms",
            "%(query_count)d queries in %(mql_time).3f ms",
            query_count,
        ) % {
            "query_count": query_count,
            "mql_time": stats.get("mql_time"),
        }

    def record(self, **kwargs):
        kwargs["djdt_query_id"] = uuid.uuid4().hex
        self._queries.append(kwargs)
        alias = kwargs["alias"]
        if alias not in self._databases:
            self._databases[alias] = {
                "time_spent": kwargs["duration"],
                "num_queries": 1,
            }
        else:
            self._databases[alias]["time_spent"] += kwargs["duration"]
            self._databases[alias]["num_queries"] += 1
        self._mql_time += kwargs["duration"]

    @property
    def title(self):
        stats = self.get_stats()
        databases = stats.get("databases", {}) if stats else {}
        count = len(databases)
        return ngettext(
            "MQL queries from %(count)d connection",
            "MQL queries from %(count)d connections",
            count,
        ) % {"count": count}



from debug_toolbar.toolbar import DebugToolbar
from django.contrib.auth.models import User
from django.http import HttpResponse
from django.test import RequestFactory, TestCase, override_settings

from django_mongodb_extensions.mql_panel import MQLPanel

request_factory = RequestFactory()


def mql_call():
    """Execute a MongoDB query for testing."""
    return list(User.objects.all())


class MQLPanelTests(TestCase):
    panel_id = MQLPanel.panel_id

    def setUp(self):
        self._get_response = lambda request: HttpResponse()
        self.request = request_factory.get("/")
        self.toolbar = DebugToolbar(self.request, self.get_response)
        self.toolbar.stats = {}
        self.panel = self.toolbar.get_panel_by_id(self.panel_id)
        self.panel.enable_instrumentation()

    def tearDown(self):
        self.panel.disable_instrumentation()

    def get_response(self, request):
        return self._get_response(request)

    def test_disabled(self):
        config = {"DISABLE_PANELS": {"django_mongodb_extensions.mql_panel.MQLPanel"}}
        self.assertIs(self.panel.enabled, True)
        with self.settings(DEBUG_TOOLBAR_CONFIG=config):
            self.assertIs(self.panel.enabled, False)

    def test_recording(self):
        self.assertEqual(len(self.panel._queries), 0)
        mql_call()
        self.assertEqual(len(self.panel._queries), 1)
        query = self.panel._queries[0]
        self.assertEqual(query["alias"], "default")
        self.assertEqual(query["mql"], "db.auth_user.aggregate([])")
        self.assertIsInstance(query["duration"], (int, float))
        self.assertGreaterEqual(query["duration"], 0)
        self.assertIsInstance(query["stacktrace"], list)
        self.assertGreater(len(query["stacktrace"]), 0)

    def test_generate_server_timing(self):
        self.assertEqual(len(self.panel._queries), 0)
        mql_call()
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.panel.generate_server_timing(self.request, response)
        self.assertEqual(len(self.panel._queries), 1)
        query = self.panel._queries[0]
        expected_data = {
            "mql_time": {"title": "MQL 1 queries", "value": query["duration"]}
        }
        self.assertEqual(self.panel.get_server_timing_stats(), expected_data)

    def test_non_ascii_query(self):
        list(User.objects.filter(username="thé"))
        list(User.objects.filter(username="café"))
        self.assertEqual(len(self.panel._queries), 2)
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertIn("café", self.panel.content)

    @override_settings(DEBUG_TOOLBAR_CONFIG={"ENABLE_STACKTRACES": False})
    def test_disable_stacktraces(self):
        mql_call()
        self.assertEqual(self.panel._queries[0]["stacktrace"], [])

    def test_duplicate_grouping(self):
        """Identical queries are marked with duplicate_count and a shared color."""
        User.objects.filter(username="user1").count()
        User.objects.filter(username="user1").count()
        User.objects.filter(username="user2").count()
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertEqual(len(self.panel._queries), 3)
        queries = self.panel._queries
        # First two queries are duplicates (identical MQL string).
        self.assertEqual(queries[0]["duplicate_count"], 2)
        self.assertEqual(queries[1]["duplicate_count"], 2)
        self.assertEqual(queries[0]["duplicate_color"], queries[1]["duplicate_color"])
        # Third query has different params so is not a duplicate.
        self.assertNotIn("duplicate_count", queries[2])

    def test_has_content_property(self):
        self.assertIs(self.panel.has_content, False)
        mql_call()
        self.assertIs(self.panel.has_content, True)

    def test_nav_subtitle(self):
        mql_call()
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertRegex(self.panel.nav_subtitle, r"1 query in \d+\.\d+ ms")

    def test_title(self):
        mql_call()
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertEqual(self.panel.title, "MQL queries from 1 connection")

    def test_slow_query_marking(self):
        mql_call()
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertIs(self.panel._queries[0]["is_slow"], False)

    @override_settings(DJDT_MQL_WARNING_THRESHOLD=0)
    def test_slow_query_marking_custom_threshold(self):
        mql_call()
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertIs(self.panel._queries[0]["is_slow"], True)

    def test_database_tracking(self):
        mql_call()
        mql_call()
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertEqual(len(self.panel._queries), 2)
        self.assertIn("default", self.panel._databases)
        db_stats = self.panel._databases["default"]
        self.assertEqual(db_stats["num_queries"], 2)
        self.assertGreater(db_stats["time_spent"], 0)

    def test_query_width_ratio(self):
        mql_call()
        mql_call()
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        total_width = sum(q["width_ratio"] for q in self.panel._queries)
        self.assertAlmostEqual(total_width, 100, places=5)



from debug_toolbar.forms import SignedDataForm
from debug_toolbar.toolbar import DebugToolbar
from django.contrib.auth.models import User
from django.http import HttpResponse
from django.test import RequestFactory, SimpleTestCase, TestCase
from django.urls import reverse

from django_mongodb_extensions.mql_panel import MQLPanel

request_factory = RequestFactory()


class MQLQueryViewTests(SimpleTestCase):
    def test_invalid_signature(self):
        url = reverse("djdt:mql_query")
        response = self.client.post(url, {"signed": "invalid"})
        self.assertContains(response, "Invalid signature", status_code=400)


class MQLExplainViewTests(SimpleTestCase):
    def test_invalid_signature(self):
        url = reverse("djdt:mql_explain")
        response = self.client.post(url, {"signed": "invalid"})
        self.assertContains(response, "Invalid signature", status_code=400)


class MQLViewDataMixin:
    def setUp(self):
        request = request_factory.get("/")
        self.toolbar = DebugToolbar(request, lambda r: HttpResponse())
        self.toolbar.stats = {}
        self.panel = self.toolbar.get_panel_by_id(MQLPanel.panel_id)
        self.panel.enable_instrumentation()
        list(User.objects.all())
        response = self.panel.process_request(request)
        self.panel.generate_stats(request, response)
        self.query = self.panel._queries[0]

    def tearDown(self):
        self.panel.disable_instrumentation()

    def _post_data(self):
        return {
            "signed": SignedDataForm.sign(
                {
                    "djdt_query_id": self.query["djdt_query_id"],
                    "request_id": self.toolbar.request_id,
                }
            )
        }


class MQLQueryViewDataTests(MQLViewDataMixin, TestCase):
    def test_renders_query_results_header(self):
        response = self.client.post(reverse("djdt:mql_query"), self._post_data())
        self.assertContains(response, "Query Results")

    def test_renders_executed_mql(self):
        response = self.client.post(reverse("djdt:mql_query"), self._post_data())
        self.assertContains(response, "Executed MQL")

    def test_renders_database_alias(self):
        response = self.client.post(reverse("djdt:mql_query"), self._post_data())
        self.assertContains(response, "default")

    def test_renders_empty_set(self):
        response = self.client.post(reverse("djdt:mql_query"), self._post_data())
        self.assertContains(response, "Empty set")


class MQLExplainViewDataTests(MQLViewDataMixin, TestCase):
    def test_renders_mql_explained_header(self):
        response = self.client.post(reverse("djdt:mql_explain"), self._post_data())
        self.assertContains(response, "Explain Results")

    def test_renders_executed_mql(self):
        response = self.client.post(reverse("djdt:mql_explain"), self._post_data())
        self.assertContains(response, "Executed MQL")

    def test_renders_explain_output(self):
        response = self.client.post(reverse("djdt:mql_explain"), self._post_data())
        self.assertContains(response, "Explain Results")



import json

from bson import json_util
from debug_toolbar.panels.sql.forms import SQLSelectForm
from debug_toolbar.toolbar import DebugToolbar
from django import forms
from django.core.exceptions import ValidationError
from django.db import connections
from django.utils.translation import gettext_lazy as _

from .utils import (
    get_max_query_results,
    parse_query_args,
)


class MQLBaseForm(SQLSelectForm):
    def _execute_operation(self, operation_type, executor_func):
        query_dict = self.cleaned_data["query"]
        alias = query_dict.get("alias", "default")
        connection = connections[alias]
        db = connection.database
        collection_name, operation, args_list = parse_query_args(query_dict)
        collection = db[collection_name]
        return executor_func(
            db,
            collection,
            collection_name,
            operation,
            args_list,
        )

    def clean(self):
        from .panel import MQLPanel

        # Call forms.Form.clean() to bypass SQLSelectForm.clean() which has
        # SQL-specific validation
        cleaned_data = forms.Form.clean(self)
        request_id = cleaned_data.get("request_id")
        if not request_id:
            raise ValidationError(_("Missing request ID."))
        djdt_query_id = cleaned_data.get("djdt_query_id")
        if not djdt_query_id:
            raise ValidationError(_("Missing query ID."))
        toolbar = DebugToolbar.fetch(request_id, panel_id=MQLPanel.panel_id)
        if toolbar is None:
            raise ValidationError(_("Data for this panel isn't available anymore."))
        panel = toolbar.get_panel_by_id(MQLPanel.panel_id)
        stats = panel.get_stats()
        if not stats or "queries" not in stats:
            raise ValidationError(_("Query data is not available."))
        # Find query in stats using djdt_query_id
        query = None
        for _query in stats["queries"]:
            if (
                isinstance(_query, dict)
                and _query.get("djdt_query_id") == djdt_query_id
            ):
                query = _query
                break
        if not query:
            raise ValidationError(_("Invalid query ID."))
        # Ensure query contains required keys
        if not all(key in query for key in ("alias", "mql")):
            raise ValidationError(_("Query data is incomplete."))
        cleaned_data["query"] = query
        return cleaned_data


class MQLExplainForm(MQLBaseForm):
    def _execute_aggregate(self, db, collection_name, args_list):
        pipeline = args_list[0] if args_list else []
        return db.command(
            "explain",
            {"aggregate": collection_name, "pipeline": pipeline, "cursor": {}},
        )

    def _execute_explain(self, db, collection, collection_name, operation, args_list):
        if operation == "aggregate":
            explain_result = self._execute_aggregate(db, collection_name, args_list)
        else:
            raise ValueError(f"Unsupported operation: {operation}")
        explain_json = json_util.dumps(explain_result, indent=4)
        result = [[explain_json]]
        headers = ["MongoDB Explain Output (JSON)"]
        return result, headers

    def explain(self):
        return self._execute_operation("explain", self._execute_explain)


class MQLQueryForm(MQLBaseForm):
    def _execute_aggregate(self, collection, args_list):
        pipeline = args_list[0] if args_list else []
        result_docs = []
        max_results = get_max_query_results()
        with collection.aggregate(pipeline) as cursor:
            for i, doc in enumerate(cursor):
                if i >= max_results:
                    break
                result_docs.append(doc)
        return result_docs

    def _execute_query(self, db, collection, collection_name, operation, args_list):
        if operation == "aggregate":
            result_docs = self._execute_aggregate(collection, args_list)
        else:
            raise ValueError(f"Unsupported read operation: {operation}")
        return self.convert_documents_to_table(result_docs)

    def _flatten_single_key_dicts(self, obj):
        if isinstance(obj, dict):
            if len(obj) == 1 and next(iter(obj)).startswith("$"):
                only_value = next(iter(obj.values()))
                return self._flatten_single_key_dicts(only_value)
            return {
                key_name: self._flatten_single_key_dicts(value_item)
                for key_name, value_item in obj.items()
            }
        elif isinstance(obj, list):
            return [self._flatten_single_key_dicts(value_item) for value_item in obj]
        return obj

    def _format_cell_value(self, value):
        serialized = json_util.dumps(value)
        parsed_json = json.loads(serialized)
        flattened_value = self._flatten_single_key_dicts(parsed_json)
        if isinstance(flattened_value, (str, int, float, bool)):
            return {"value": str(flattened_value), "is_json": False}
        if isinstance(flattened_value, dict):
            return {
                "type": "dict",
                "value": self._format_dict_for_template(flattened_value),
                "is_json": False,
            }
        if isinstance(flattened_value, list):
            return {
                "type": "list",
                "value": self._format_list_for_template(flattened_value),
                "is_json": False,
            }
        return {
            "value": json.dumps(flattened_value, indent=4),
            "is_json": True,
        }

    def _format_dict_for_template(self, dictionary):
        return [
            {"key": key_name, **self._format_cell_value(value_item)}
            for key_name, value_item in dictionary.items()
        ]

    def _format_list_for_template(self, list_items):
        return [
            {"key": index, **self._format_cell_value(value_item)}
            for index, value_item in enumerate(list_items)
        ]

    def _format_row(self, row_dict):
        return [self._format_cell_value(cell_value) for cell_value in row_dict.values()]

    def convert_documents_to_table(self, documents):
        if not documents:
            return [], []
        # Collect all unique field names
        all_fields = set()
        for doc in documents:
            all_fields.update(doc.keys())
        # Sort fields for consistent column ordering, with _id first if present
        headers = sorted(all_fields)
        if "_id" in headers:
            headers.remove("_id")
            headers.insert(0, "_id")
        # Convert each document to a row with formatted values
        rows = []
        for doc in documents:
            row = [self._format_cell_value(doc.get(field)) for field in headers]
            rows.append(row)
        return rows, headers

    def query(self):
        return self._execute_operation("query", self._execute_query)



import types
import weakref

from bson import json_util
from debug_toolbar.utils import get_stack_trace, get_template_info
from django.conf import settings
from django_mongodb_backend.utils import OperationDebugWrapper

DEFAULT_MAX_QUERY_RESULTS = 100
DEFAULT_MQL_WARNING_THRESHOLD = 500
_patched_connections = weakref.WeakSet()


class DebugToolbarWrapper(OperationDebugWrapper):
    """Wrapper around pymongo Collection objects that logs queries
    for display and re-execution."""

    def __init__(self, db, collection, logger):
        super().__init__(db, collection)
        self.logger = logger

    def log(self, op, duration, args, kwargs=None):
        args_str = ", ".join(repr(arg) for arg in args)
        operation = f"db.{self.collection_name}{op}({args_str})"
        args_json = json_util.dumps(list(args))
        if self.logger:
            self.logger.record(
                alias=self.db.alias,
                mql=operation,
                duration=duration,
                stacktrace=get_stack_trace(),
                template_info=get_template_info(),
                mql_collection=self.collection_name.strip("."),
                mql_operation=op.lstrip("."),
                mql_args_json=args_json,
            )


def format_mql_query(query):
    """Return a pretty-printed MQL query string."""
    collection_name, operation, args_list = parse_query_args(query)
    if args_list:
        args_formatted = json_util.dumps(
            args_list[0] if len(args_list) == 1 else args_list,
            indent=4,
        )
    else:
        args_formatted = ""
    return f"db.{collection_name}.{operation}(\n{args_formatted}\n)"


def get_max_query_results():
    """Get the maximum number of results to return when re-executing queries.

    Return the value from Django settings DJDT_MQL_MAX_QUERY_RESULTS if set,
    otherwise return the default value.
    """
    return getattr(settings, "DJDT_MQL_MAX_QUERY_RESULTS", DEFAULT_MAX_QUERY_RESULTS)


def get_mql_warning_threshold():
    """Get the slow-query warning threshold in milliseconds.

    Return the value from Django settings DJDT_MQL_WARNING_THRESHOLD if set,
    otherwise return the default value.
    """
    return getattr(
        settings, "DJDT_MQL_WARNING_THRESHOLD", DEFAULT_MQL_WARNING_THRESHOLD
    )


def parse_query_args(query_dict):
    """Extract and deserialize collection name, operation, and arguments."""
    collection_name = query_dict["mql_collection"]
    operation = query_dict["mql_operation"]
    args_json = query_dict["mql_args_json"]
    return collection_name, operation, json_util.loads(args_json)


def patch_get_collection(connection):
    """Patch get_collection to log queries via DebugToolbarWrapper."""
    if connection in _patched_connections:
        return
    if not hasattr(connection, "_original_get_collection"):
        connection._original_get_collection = connection.get_collection

    def get_collection(self, name, **kwargs):
        # Use _mql_djdt_logger (not _djdt_logger) to avoid conflicting with the
        # SQL panel, which sets _djdt_logger on all connections.
        logger = getattr(self, "_mql_djdt_logger", None)
        if logger:
            collection = self._original_get_collection(name, **kwargs)
            return DebugToolbarWrapper(self, collection, logger)
        else:
            return self._original_get_collection(name, **kwargs)

    connection.get_collection = types.MethodType(get_collection, connection)
    _patched_connections.add(connection)


def patch_new_connection(sender, connection, **kwargs):
    """Handle connection_created by patching new MongoDB connections."""
    if hasattr(connection, "database") and hasattr(connection, "get_collection"):
        patch_get_collection(connection)



from django.contrib.admin.apps import AdminConfig
from django.contrib.auth.apps import AuthConfig
from django.contrib.contenttypes.apps import ContentTypesConfig


class MongoAdminConfig(AdminConfig):
    default_auto_field = "django_mongodb_backend.fields.ObjectIdAutoField"


class MongoAuthConfig(AuthConfig):
    default_auto_field = "django_mongodb_backend.fields.ObjectIdAutoField"


class MongoContentTypesConfig(ContentTypesConfig):
    default_auto_field = "django_mongodb_backend.fields.ObjectIdAutoField"



import os


def always_show_toolbar(request):
    return True


SECRET_KEY = "test-secret-key-for-django-mongodb-extensions"
DEBUG = True
INSTALLED_APPS = [
    "tests.apps.MongoAuthConfig",
    "tests.apps.MongoContentTypesConfig",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "debug_toolbar",
    "django_mongodb_extensions",
]
DATABASES = {
    "default": {
        "ENGINE": "django_mongodb_backend",
        "NAME": "django_mongodb_extensions",
        "HOST": os.environ.get("MONGODB_URI", "mongodb://localhost:27017/"),
    }
}
DATABASE_ROUTERS = ["django_mongodb_backend.routers.MongoRouter"]
MIDDLEWARE = []
# Use a callback that doesn't check DEBUG, since tests run with DEBUG=False.
DEBUG_TOOLBAR_CONFIG = {
    "SHOW_TOOLBAR_CALLBACK": always_show_toolbar,
}
ROOT_URLCONF = "tests.urls"
USE_TZ = True
STATIC_URL = "/static/"
DEFAULT_AUTO_FIELD = "django_mongodb_backend.fields.ObjectIdAutoField"
# Disable migrations for tests
MIGRATION_MODULES = {
    "auth": None,
    "contenttypes": None,
    "sessions": None,
}
# Include MQLPanel for testing
DEBUG_TOOLBAR_PANELS = [
    "debug_toolbar.panels.history.HistoryPanel",
    "debug_toolbar.panels.versions.VersionsPanel",
    "debug_toolbar.panels.timer.TimerPanel",
    "debug_toolbar.panels.settings.SettingsPanel",
    "debug_toolbar.panels.headers.HeadersPanel",
    "debug_toolbar.panels.request.RequestPanel",
    "debug_toolbar.panels.sql.SQLPanel",
    "django_mongodb_extensions.mql_panel.MQLPanel",
    "debug_toolbar.panels.staticfiles.StaticFilesPanel",
    "debug_toolbar.panels.templates.TemplatesPanel",
    "debug_toolbar.panels.alerts.AlertsPanel",
    "debug_toolbar.panels.cache.CachePanel",
    "debug_toolbar.panels.signals.SignalsPanel",
    "debug_toolbar.panels.redirects.RedirectsPanel",
    "debug_toolbar.panels.profiling.ProfilingPanel",
]
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
            ],
        },
    },
]



from debug_toolbar.decorators import render_with_toolbar_language, require_show_toolbar
from debug_toolbar.panels.sql.views import get_signed_data
from django.contrib.auth.decorators import login_not_required
from django.http import HttpResponseBadRequest, JsonResponse
from django.template.loader import render_to_string
from django.views.decorators.csrf import csrf_exempt

from .forms import MQLExplainForm, MQLQueryForm
from .utils import format_mql_query


@csrf_exempt
@login_not_required
@require_show_toolbar
@render_with_toolbar_language
def mql_explain(request):
    verified_data = get_signed_data(request)
    if not verified_data:
        return HttpResponseBadRequest("Invalid signature")
    form = MQLExplainForm(verified_data)
    if form.is_valid():
        query = form.cleaned_data["query"]
        result, headers = form.explain()
        context = {
            "result": result,
            "mql": format_mql_query(query),
            "duration": query["duration"],
            "headers": headers,
            "alias": query["alias"],
        }
        content = render_to_string("mql_panel/mql_explain.html", context)
        return JsonResponse({"content": content})
    return HttpResponseBadRequest("Form errors")


@csrf_exempt
@login_not_required
@require_show_toolbar
@render_with_toolbar_language
def mql_query(request):
    verified_data = get_signed_data(request)
    if not verified_data:
        return HttpResponseBadRequest("Invalid signature")
    form = MQLQueryForm(verified_data)
    if form.is_valid():
        query = form.cleaned_data["query"]
        result, headers = form.query()

        context = {
            "result": result,
            "mql": format_mql_query(query),
            "duration": query["duration"],
            "headers": headers,
            "alias": query["alias"],
        }
        content = render_to_string("mql_panel/mql_query.html", context)
        return JsonResponse({"content": content})
    return HttpResponseBadRequest("Form errors")
